"""Functions for converting tree data into data suitable for the FilterSet."""

from typing import Any, Dict, List, Type, Union

from django.contrib.postgres.search import (
    SearchQuery,
    SearchRank,
    SearchVector,
    TrigramDistance,
    TrigramSimilarity,
)
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django_filters.conf import settings as django_settings
from graphene.types.inputobjecttype import InputObjectTypeContainer
from graphene_django_filter.filters import SearchQueryFilter, SearchRankFilter, TrigramFilter
from graphene_django_filter.input_types import (
    SearchConfigInputType,
    SearchQueryFilterInputType,
    SearchQueryInputType,
    SearchRankFilterInputType,
    SearchRankWeightsInputType,
    SearchVectorInputType,
    TrigramFilterInputType,
    TrigramSearchKind,
)

from .conf import settings
from .filterset import AdvancedFilterSet


def tree_input_type_to_data(
    filterset_class: Type[AdvancedFilterSet],
    tree_input_type: InputObjectTypeContainer,
    prefix: str = '',
) -> Dict[str, Any]:
    """Convert a tree_input_type to a FilterSet data."""
    result: Dict[str, Any] = {}
    for key, value in tree_input_type.items():
        if key in ('and', 'or'):
            result[key] = [tree_input_type_to_data(filterset_class, subtree) for subtree in value]
        elif key == 'not':
            result[key] = tree_input_type_to_data(filterset_class, value)
        else:
            result.update(
                create_data(
                    (prefix + LOOKUP_SEP + key if prefix else key).replace(
                        LOOKUP_SEP + django_settings.DEFAULT_LOOKUP_EXPR, '',
                    ),
                    value,
                    filterset_class,
                ),
            )
    return result


def create_data(key: str, value: Any, filterset_class: Type[AdvancedFilterSet]) -> Dict[str, Any]:
    """Create data from a key and a value."""
    for factory_key, factory in DATA_FACTORIES.items():
        if factory_key in key:
            return factory(value, key, filterset_class)
    if isinstance(value, InputObjectTypeContainer):
        return tree_input_type_to_data(filterset_class, value, key)
    else:
        return {key: value}


def create_search_query_data(
    input_type: SearchQueryFilterInputType,
    key: str,
    filterset_class: Type[AdvancedFilterSet],
) -> Dict[str, SearchQueryFilter.Value]:
    """Create a data for the `SearchQueryFilter` class."""
    return {
        key: SearchQueryFilter.Value(
            annotation_value=create_search_vector(input_type.vector, filterset_class),
            search_value=create_search_query(input_type.query),
        ),
    }


def create_search_rank_data(
    input_type: Union[SearchRankFilterInputType, InputObjectTypeContainer],
    key: str,
    filterset_class: Type[AdvancedFilterSet],
) -> Dict[str, SearchRankFilter.Value]:
    """Create a data for the `SearchRankFilter` class."""
    rank_data = {}
    for lookup, value in input_type.lookups.items():
        search_rank_data = {
            'vector': create_search_vector(input_type.vector, filterset_class),
            'query': create_search_query(input_type.query),
            'cover_density': input_type.cover_density,
        }
        weights = input_type.get('weights', None)
        if weights:
            search_rank_data['weights'] = create_search_rank_weights(weights)
        normalization = input_type.get('normalization', None)
        if normalization:
            search_rank_data['normalization'] = normalization
        k = (key + LOOKUP_SEP + lookup).replace(
            LOOKUP_SEP + django_settings.DEFAULT_LOOKUP_EXPR, '',
        )
        rank_data[k] = SearchRankFilter.Value(
            annotation_value=SearchRank(**search_rank_data),
            search_value=value,
        )
    return rank_data


def create_trigram_data(
    input_type: TrigramFilterInputType,
    key: str,
    *args
) -> Dict[str, TrigramFilter.Value]:
    """Create a data for the `TrigramFilter` class."""
    trigram_data = {}
    if input_type.kind == TrigramSearchKind.SIMILARITY:
        trigram_class = TrigramSimilarity
    else:
        trigram_class = TrigramDistance
    for lookup, value in input_type.lookups.items():
        k = (key + LOOKUP_SEP + lookup).replace(
            LOOKUP_SEP + django_settings.DEFAULT_LOOKUP_EXPR, '',
        )
        trigram_data[k] = TrigramFilter.Value(
            annotation_value=trigram_class(
                LOOKUP_SEP.join(key.split(LOOKUP_SEP)[:-1]),
                input_type.value,
            ),
            search_value=value,
        )
    return trigram_data


def create_search_vector(
    input_type: Union[SearchVectorInputType, InputObjectTypeContainer],
    filterset_class: Type[AdvancedFilterSet],
) -> SearchVector:
    """Create an object of the `SearchVector` class."""
    validate_search_vector_fields(filterset_class, input_type.fields)
    search_vector_data = {}
    config = input_type.get('config', None)
    if config:
        search_vector_data['config'] = create_search_config(config)
    weight = input_type.get('weight', None)
    if weight:
        search_vector_data['weight'] = weight.value
    return SearchVector(*input_type.fields, **search_vector_data)


def create_search_query(
    input_type: Union[SearchQueryInputType, InputObjectTypeContainer],
) -> SearchQuery:
    """Create an object of the `SearchQuery` class."""
    validate_search_query(input_type)
    value = input_type.get('value', None)
    if value:
        config = input_type.get('config', None)
        if config:
            search_query = SearchQuery(input_type.value, config=create_search_config(config))
        else:
            search_query = SearchQuery(input_type.value)
    else:
        search_query = None
    and_search_query = None
    for and_input_type in input_type.get(settings.AND_KEY, []):
        if and_search_query is None:
            and_search_query = create_search_query(and_input_type)
        else:
            and_search_query = and_search_query & create_search_query(and_input_type)
    or_search_query = None
    for or_input_type in input_type.get(settings.OR_KEY, []):
        if or_search_query is None:
            or_search_query = create_search_query(or_input_type)
        else:
            or_search_query = or_search_query | create_search_query(or_input_type)
    not_input_type = input_type.get(settings.NOT_KEY, None)
    not_search_query = create_search_query(not_input_type) if not_input_type else None
    valid_queries = (
        q for q in (and_search_query, or_search_query, not_search_query) if q is not None
    )
    for valid_query in valid_queries:
        search_query = search_query & valid_query if search_query else valid_query
    return search_query


def create_search_config(input_type: SearchConfigInputType) -> Union[str, models.F]:
    """Create a `SearchVector` or `SearchQuery` object config."""
    return models.F(input_type.value) if input_type.is_field else input_type.value


def create_search_rank_weights(input_type: SearchRankWeightsInputType) -> List[float]:
    """Create a search rank weights list."""
    return [input_type.D, input_type.C, input_type.B, input_type.A]


def validate_search_vector_fields(
    filterset_class: Type[AdvancedFilterSet],
    fields: List[str],
) -> None:
    """Validate that fields is included in full text search fields."""
    full_text_search_fields = filterset_class.get_full_text_search_fields()
    for field in fields:
        if field not in full_text_search_fields:
            raise ValidationError(f'The `{field}` field is not included in full text search fields')


def validate_search_query(
    input_type: Union[SearchQueryInputType, InputObjectTypeContainer],
) -> None:
    """Validate that search query contains at least one required field."""
    if all([
        'value' not in input_type,
        settings.AND_KEY not in input_type,
        settings.OR_KEY not in input_type,
        settings.NOT_KEY not in input_type,
    ]):
        raise ValidationError(
            'The search query must contains at least one required field '
            f'such as `value`, `{settings.AND_KEY}`, `{settings.OR_KEY}`, `{settings.NOT_KEY}`.',
        )


DATA_FACTORIES = {
    SearchQueryFilter.postfix: create_search_query_data,
    SearchRankFilter.postfix: create_search_rank_data,
    TrigramFilter.postfix: create_trigram_data,
}



"""Functions for creating a FilterSet class."""

from typing import Optional, Type

from graphene_django.filter.filterset import custom_filterset_factory, setup_filterset
from graphene_django.filter.utils import replace_csv_filters

from .filterset import AdvancedFilterSet


def get_filterset_class(
    filterset_class: Optional[Type[AdvancedFilterSet]],
    **meta
) -> Type[AdvancedFilterSet]:
    """Get a class to be used as a FilterSet.

    It is a partial copy of the `get_filterset_class` function from graphene-django.
    https://github.com/graphql-python/graphene-django/blob/caf954861025b9f3d9d3f9c204a7cbbc87352265/graphene_django/filter/utils.py#L56
    """
    if filterset_class:
        graphene_filterset_class = setup_filterset(filterset_class)
    else:
        graphene_filterset_class = custom_filterset_factory(
            filterset_base_class=AdvancedFilterSet,
            **meta
        )
    replace_csv_filters(graphene_filterset_class)
    return graphene_filterset_class



"""Library settings."""

from typing import Any, Dict, Optional, Union

from django.conf import settings as django_settings
from django.db import connection
from django.test.signals import setting_changed


def get_fixed_settings() -> Dict[str, bool]:
    """Return fixed settings."""
    is_postgresql = connection.vendor == 'postgresql'
    has_trigram_extension = False
    if is_postgresql:
        with connection.cursor() as cursor:
            cursor.execute("SELECT COUNT(*) FROM pg_available_extensions WHERE name='pg_trgm'")
            has_trigram_extension = cursor.fetchone()[0] == 1
    return {
        'IS_POSTGRESQL': is_postgresql,
        'HAS_TRIGRAM_EXTENSION': has_trigram_extension,
    }


FIXED_SETTINGS = get_fixed_settings()
DEFAULT_SETTINGS = {
    'FILTER_KEY': 'filter',
    'AND_KEY': 'and',
    'OR_KEY': 'or',
    'NOT_KEY': 'not',
}
DJANGO_SETTINGS_KEY = 'GRAPHENE_DJANGO_FILTER'


class Settings:
    """Library settings.

    Settings consist of fixed ones that depend on the user environment
    and others that can be set with Django settings.py module.
    """

    def __init__(self, user_settings: Optional[dict] = None) -> None:
        self._user_settings = user_settings

    @property
    def user_settings(self) -> dict:
        """Return user-defined settings."""
        if self._user_settings is None:
            self._user_settings = getattr(django_settings, DJANGO_SETTINGS_KEY, {})
        return self._user_settings

    def __getattr__(self, name: str) -> Union[str, bool]:
        """Return a setting value."""
        if name not in FIXED_SETTINGS and name not in DEFAULT_SETTINGS:
            raise AttributeError(f'Invalid Graphene setting: `{name}`')
        if name in FIXED_SETTINGS:
            return FIXED_SETTINGS[name]
        elif name in self.user_settings:
            return self.user_settings[name]
        else:
            return DEFAULT_SETTINGS[name]


settings = Settings(None)


def reload_settings(setting: str, value: Any, **kwargs) -> None:
    """Reload settings in response to the `setting_changed` signal."""
    global settings
    if setting == DJANGO_SETTINGS_KEY:
        settings = Settings(value)


setting_changed.connect(reload_settings)



"""Additional filters for special lookups."""

from typing import Any, Callable, NamedTuple, Optional, Union

from django.contrib.postgres.search import (
    SearchQuery,
    SearchRank,
    SearchVector,
    TrigramDistance,
    TrigramSimilarity,
)
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django_filters import Filter
from django_filters.constants import EMPTY_VALUES


class AnnotatedFilter(Filter):
    """Filter with a QuerySet object annotation."""

    class Value(NamedTuple):
        annotation_value: Any
        search_value: Any

    postfix = 'annotated'

    def __init__(
        self,
        field_name: Optional[str] = None,
        lookup_expr: Optional[str] = None,
        *,
        label: Optional[str] = None,
        method: Optional[Union[str, Callable]] = None,
        distinct: bool = False,
        exclude: bool = False,
        **kwargs
    ) -> None:
        super().__init__(
            field_name,
            lookup_expr,
            label=label,
            method=method,
            distinct=distinct,
            exclude=exclude,
            **kwargs
        )
        self.filter_counter = 0

    @property
    def annotation_name(self) -> str:
        """Return the name used for the annotation."""
        return f'{self.field_name}_{self.postfix}_{self.creation_counter}_{self.filter_counter}'

    def filter(self, qs: models.QuerySet, value: Value) -> models.QuerySet:
        """Filter a QuerySet using annotation."""
        if value in EMPTY_VALUES:
            return qs
        if self.distinct:
            qs = qs.distinct()
        annotation_name = self.annotation_name
        self.filter_counter += 1
        qs = qs.annotate(**{annotation_name: value.annotation_value})
        lookup = f'{annotation_name}{LOOKUP_SEP}{self.lookup_expr}'
        return self.get_method(qs)(**{lookup: value.search_value})


class SearchQueryFilter(AnnotatedFilter):
    """Full text search filter using the `SearchVector` and `SearchQuery` object."""

    class Value(NamedTuple):
        annotation_value: SearchVector
        search_value: SearchQuery

    postfix = 'search_query'
    available_lookups = ('exact',)

    def filter(self, qs: models.QuerySet, value: Value) -> models.QuerySet:
        """Filter a QuerySet using the `SearchVector` and `SearchQuery` object."""
        return super().filter(qs, value)


class SearchRankFilter(AnnotatedFilter):
    """Full text search filter using the `SearchRank` object."""

    class Value(NamedTuple):
        annotation_value: SearchRank
        search_value: float

    postfix = 'search_rank'
    available_lookups = ('exact', 'gt', 'gte', 'lt', 'lte')

    def filter(self, qs: models.QuerySet, value: Value) -> models.QuerySet:
        """Filter a QuerySet using the `SearchRank` object."""
        return super().filter(qs, value)


class TrigramFilter(AnnotatedFilter):
    """Full text search filter using similarity or distance of trigram."""

    class Value(NamedTuple):
        annotation_value: Union[TrigramSimilarity, TrigramDistance]
        search_value: float

    postfix = 'trigram'
    available_lookups = ('exact', 'gt', 'gte', 'lt', 'lte')

    def filter(self, qs: models.QuerySet, value: Value) -> models.QuerySet:
        """Filter a QuerySet using similarity or distance of trigram."""
        return super().filter(qs, value)



"""`AdvancedFilterSet` class module.

Use the `AdvancedFilterSet` class from this module instead of the `FilterSet` from django-filter.
"""

import warnings
from collections import OrderedDict
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Type, Union, cast

from django.db import connection, models
from django.db.models.constants import LOOKUP_SEP
from django.forms import Form
from django.forms.utils import ErrorDict
from django_filters import Filter
from django_filters.conf import settings as django_settings
from django_filters.filterset import BaseFilterSet, FilterSetMetaclass
from wrapt import ObjectProxy

from .conf import settings


class QuerySetProxy(ObjectProxy):
    """Proxy for a QuerySet object.

    The Django-filter library works with QuerySet objects,
    but such objects do not provide the ability to apply the negation operator to the entire object.
    Therefore, it is convenient to work with the Q object instead of the QuerySet.
    This class replaces the original QuerySet object,
    and creates a Q object when calling the `filter` and `exclude` methods.
    """

    __slots__ = 'q'

    def __init__(self, wrapped: models.QuerySet, q: Optional[models.Q] = None) -> None:
        super().__init__(wrapped)
        self.q = q or models.Q()

    def __getattr__(self, name: str) -> Any:
        """Return QuerySet attributes for all cases except `filter` and `exclude`."""
        if name == 'filter':
            return self.filter_
        elif name == 'exclude':
            return self.exclude_
        attr = super().__getattr__(name)
        if callable(attr):
            def func(*args, **kwargs) -> Any:
                result = attr(*args, **kwargs)
                if isinstance(result, models.QuerySet):
                    return QuerySetProxy(result, self.q)
                return result
            return func
        return attr

    def __iter__(self) -> Iterator[Any]:
        """Return QuerySet and Q objects."""
        return iter([self.__wrapped__, self.q])

    def filter_(self, *args, **kwargs) -> 'QuerySetProxy':
        """Replace the `filter` method of the QuerySet class."""
        if len(kwargs) == 0 and len(args) == 1 and isinstance(args[0], models.Q):
            q = args[0]
        else:
            q = models.Q(*args, **kwargs)
        self.q = self.q & q
        return self

    def exclude_(self, *args, **kwargs) -> 'QuerySetProxy':
        """Replace the `exclude` method of the QuerySet class."""
        if len(kwargs) == 0 and len(args) == 1 and isinstance(args[0], models.Q):
            q = args[0]
        else:
            q = models.Q(*args, **kwargs)
        self.q = self.q & ~q
        return self


def is_full_text_search_lookup_expr(lookup_expr: str) -> bool:
    """Determine if a lookup_expr is a full text search expression."""
    return lookup_expr.split(LOOKUP_SEP)[-1] == 'full_text_search'


def is_regular_lookup_expr(lookup_expr: str) -> bool:
    """Determine whether the lookup_expr must be processed in a regular way."""
    return not any([
        is_full_text_search_lookup_expr(lookup_expr),
    ])


class AdvancedFilterSet(BaseFilterSet, metaclass=FilterSetMetaclass):
    """Allow you to use advanced filters."""

    class TreeFormMixin(Form):
        """Tree-like form mixin."""

        def __init__(
            self,
            and_forms: Optional[List['AdvancedFilterSet.TreeFormMixin']] = None,
            or_forms: Optional[List['AdvancedFilterSet.TreeFormMixin']] = None,
            not_form: Optional['AdvancedFilterSet.TreeFormMixin'] = None,
            *args,
            **kwargs
        ) -> None:
            super().__init__(*args, **kwargs)
            self.and_forms = and_forms or []
            self.or_forms = or_forms or []
            self.not_form = not_form

        @property
        def errors(self) -> ErrorDict:
            """Return an ErrorDict for the data provided for the form."""
            self_errors: ErrorDict = super().errors
            for key in ('and', 'or'):
                errors: ErrorDict = ErrorDict()
                for i, form in enumerate(getattr(self, f'{key}_forms')):
                    if form.errors:
                        errors[f'{key}_{i}'] = form.errors
                if len(errors):
                    self_errors.update({key: errors})
            if self.not_form and self.not_form.errors:
                self_errors.update({'not': self.not_form.errors})
            return self_errors

    def get_form_class(self) -> Type[Union[Form, TreeFormMixin]]:
        """Return a django Form class suitable of validating the filterset data.

        The form must be tree-like because the data is tree-like.
        """
        form_class = super(AdvancedFilterSet, self).get_form_class()
        tree_form = cast(
            Type[Union[Form, AdvancedFilterSet.TreeFormMixin]],
            type(
                f'{form_class.__name__.replace("Form", "")}TreeForm',
                (form_class, AdvancedFilterSet.TreeFormMixin),
                {},
            ),

        )
        return tree_form

    @property
    def form(self) -> Union[Form, TreeFormMixin]:
        """Return a django Form suitable of validating the filterset data."""
        if not hasattr(self, '_form'):
            form_class = self.get_form_class()
            if self.is_bound:
                self._form = self.create_form(form_class, self.data)
            else:
                self._form = form_class(prefix=self.form_prefix)
        return self._form

    def create_form(
        self,
        form_class: Type[Union[Form, TreeFormMixin]],
        data: Dict[str, Any],
    ) -> Union[Form, TreeFormMixin]:
        """Create a form from a form class and data."""
        return form_class(
            data={k: v for k, v in data.items() if k not in ('and', 'or', 'not')},
            and_forms=[self.create_form(form_class, and_data) for and_data in data.get('and', [])],
            or_forms=[self.create_form(form_class, or_data) for or_data in data.get('or', [])],
            not_form=self.create_form(form_class, data['not']) if data.get('not', None) else None,
        )

    def find_filter(self, data_key: str) -> Filter:
        """Find a filter using a data key.

        The data key may differ from a filter name, because
        the data keys may contain DEFAULT_LOOKUP_EXPR and user can create
        a AdvancedFilterSet class without following the naming convention.
        """
        if LOOKUP_SEP in data_key:
            field_name, lookup_expr = data_key.rsplit(LOOKUP_SEP, 1)
        else:
            field_name, lookup_expr = data_key, django_settings.DEFAULT_LOOKUP_EXPR
        key = field_name if lookup_expr == django_settings.DEFAULT_LOOKUP_EXPR else data_key
        if key in self.filters:
            return self.filters[key]
        for filter_value in self.filters.values():
            if filter_value.field_name == field_name and filter_value.lookup_expr == lookup_expr:
                return filter_value

    def filter_queryset(self, queryset: models.QuerySet) -> models.QuerySet:
        """Filter a queryset with a top level form's `cleaned_data`."""
        qs, q = self.get_queryset_proxy_for_form(queryset, self.form)
        return qs.filter(q)

    def get_queryset_proxy_for_form(
        self,
        queryset: models.QuerySet,
        form: Union[Form, TreeFormMixin],
    ) -> QuerySetProxy:
        """Return a `QuerySetProxy` object for a form's `cleaned_data`."""
        qs = queryset
        q = models.Q()
        for name, value in form.cleaned_data.items():
            qs, q = self.find_filter(name).filter(QuerySetProxy(qs, q), value)
        and_q = models.Q()
        for and_form in form.and_forms:
            qs, new_q = self.get_queryset_proxy_for_form(qs, and_form)
            and_q = and_q & new_q
        or_q = models.Q()
        for or_form in form.or_forms:
            qs, new_q = self.get_queryset_proxy_for_form(qs, or_form)
            or_q = or_q | new_q
        if form.not_form:
            qs, new_q = self.get_queryset_proxy_for_form(queryset, form.not_form)
            not_q = ~new_q
        else:
            not_q = models.Q()
        return QuerySetProxy(qs, q & and_q & or_q & not_q)

    @classmethod
    def get_filters(cls) -> OrderedDict:
        """Get all filters for the filterset.

        This is the combination of declared and generated filters.
        """
        filters = super().get_filters()
        if not cls._meta.model:
            return filters
        return OrderedDict([
            *filters.items(),
            *cls.create_full_text_search_filters(filters).items(),
        ])

    @classmethod
    def create_full_text_search_filters(
        cls,
        base_filters: OrderedDict,
    ) -> OrderedDict:
        """Create available full text search filters."""
        new_filters = OrderedDict()
        full_text_search_fields = cls.get_full_text_search_fields()
        if not len(full_text_search_fields):
            return new_filters
        if not settings.IS_POSTGRESQL:
            warnings.warn(
                f'Full text search is not available because the {connection.vendor} vendor is '
                'used instead of the postgresql vendor.',
            )
            return new_filters
        from .filters import SearchQueryFilter, SearchRankFilter, TrigramFilter
        new_filters = OrderedDict([
            *new_filters.items(),
            *cls.create_special_filters(base_filters, SearchQueryFilter).items(),
            *cls.create_special_filters(base_filters, SearchRankFilter).items(),
        ])
        if not settings.HAS_TRIGRAM_EXTENSION:
            warnings.warn(
                'Trigram search is not available because the `pg_trgm` extension is not installed.',
            )
            return new_filters
        for field_name in full_text_search_fields:
            new_filters = OrderedDict([
                *new_filters.items(),
                *cls.create_special_filters(base_filters, TrigramFilter, field_name).items(),
            ])
        return new_filters

    @classmethod
    def create_special_filters(
        cls,
        base_filters: OrderedDict,
        filter_class: Union[Type[Filter], Any],
        field_name: Optional[str] = None,
    ) -> OrderedDict:
        """Create special filters using a filter class and a field name."""
        new_filters = OrderedDict()
        for lookup_expr in filter_class.available_lookups:
            if field_name:
                postfix_field_name = f'{field_name}{LOOKUP_SEP}{filter_class.postfix}'
            else:
                postfix_field_name = filter_class.postfix
            filter_name = cls.get_filter_name(postfix_field_name, lookup_expr)
            if filter_name not in base_filters:
                new_filters[filter_name] = filter_class(
                    field_name=postfix_field_name,
                    lookup_expr=lookup_expr,
                )
        return new_filters

    @classmethod
    def get_fields(cls) -> OrderedDict:
        """Resolve the `Meta.fields` argument including only regular lookups."""
        return cls._get_fields(is_regular_lookup_expr)

    @classmethod
    def get_full_text_search_fields(cls) -> OrderedDict:
        """Resolve the `Meta.fields` argument including only full text search lookups."""
        return cls._get_fields(is_full_text_search_lookup_expr)

    @classmethod
    def _get_fields(cls, predicate: Callable[[str], bool]) -> OrderedDict:
        """Resolve the `Meta.fields` argument including lookups that match the predicate."""
        fields: List[Tuple[str, List[str]]] = []
        for k, v in super().get_fields().items():
            regular_field = [lookup_expr for lookup_expr in v if predicate(lookup_expr)]
            if len(regular_field):
                fields.append((k, regular_field))
        return OrderedDict(fields)



"""`AdvancedDjangoFilterConnectionField` class module.

Use the `AdvancedDjangoFilterConnectionField` class from this
module instead of the `DjangoFilterConnectionField` from graphene-django.
"""

import warnings
from typing import Any, Callable, Dict, Iterable, Optional, Type, Union

import graphene
from django.core.exceptions import ValidationError
from django.db import models
from graphene_django import DjangoObjectType
from graphene_django.filter import DjangoFilterConnectionField

from .conf import settings
from .filter_arguments_factory import FilterArgumentsFactory
from .filterset import AdvancedFilterSet
from .filterset_factories import get_filterset_class
from .input_data_factories import tree_input_type_to_data


class AdvancedDjangoFilterConnectionField(DjangoFilterConnectionField):
    """Allow you to use advanced filters provided by this library."""

    def __init__(
        self,
        type: Union[Type[DjangoObjectType], Callable[[], Type[DjangoObjectType]], str],
        fields: Optional[Dict[str, list]] = None,
        order_by: Any = None,
        extra_filter_meta: Optional[dict] = None,
        filterset_class: Optional[Type[AdvancedFilterSet]] = None,
        filter_input_type_prefix: Optional[str] = None,
        *args,
        **kwargs
    ) -> None:
        super().__init__(
            type,
            fields,
            order_by,
            extra_filter_meta,
            filterset_class,
            *args,
            **kwargs
        )
        assert self.provided_filterset_class is None or issubclass(
            self.provided_filterset_class, AdvancedFilterSet,
        ), 'Use the `AdvancedFilterSet` class with the `AdvancedDjangoFilterConnectionField`'
        self._filter_input_type_prefix = filter_input_type_prefix
        if self._filter_input_type_prefix is None and self._provided_filterset_class:
            warnings.warn(
                'The `filterset_class` argument without `filter_input_type_prefix` '
                'can result in different types with the same name in the schema.',
            )
        if self._filter_input_type_prefix is None and self.node_type._meta.filterset_class:
            warnings.warn(
                f'The `filterset_class` field of `{self.node_type.__name__}` Meta '
                'without the `filter_input_type_prefix` argument '
                'can result in different types with the same name in the schema.',
            )

    @property
    def provided_filterset_class(self) -> Optional[Type[AdvancedFilterSet]]:
        """Return a provided AdvancedFilterSet class."""
        return self._provided_filterset_class or self.node_type._meta.filterset_class

    @property
    def filter_input_type_prefix(self) -> str:
        """Return a prefix for a filter input type name."""
        if self._filter_input_type_prefix:
            return self._filter_input_type_prefix
        node_type_name = self.node_type.__name__.replace('Type', '')
        if self.provided_filterset_class:
            return f'{node_type_name}{self.provided_filterset_class.__name__}'
        else:
            return node_type_name

    @property
    def filterset_class(self) -> Type[AdvancedFilterSet]:
        """Return a AdvancedFilterSet instead of a FilterSet."""
        if not self._filterset_class:
            fields = self._fields or self.node_type._meta.filter_fields
            meta = {'model': self.model, 'fields': fields}
            if self._extra_filter_meta:
                meta.update(self._extra_filter_meta)
            self._filterset_class = get_filterset_class(self.provided_filterset_class, **meta)
        return self._filterset_class

    @property
    def filtering_args(self) -> dict:
        """Return filtering args from the filterset."""
        if not self._filtering_args:
            self._filtering_args = FilterArgumentsFactory(
                self.filterset_class,
                self.filter_input_type_prefix,
            ).arguments
        return self._filtering_args

    @classmethod
    def resolve_queryset(
        cls,
        connection: object,
        iterable: Iterable,
        info: graphene.ResolveInfo,
        args: Dict[str, Any],
        filtering_args: Dict[str, graphene.InputField],
        filterset_class: Type[AdvancedFilterSet],
    ) -> models.QuerySet:
        """Return a filtered QuerySet."""
        qs = super(DjangoFilterConnectionField, cls).resolve_queryset(
            connection, iterable, info, args,
        )
        filter_arg = args.get(settings.FILTER_KEY, {})
        filterset = filterset_class(
            data=tree_input_type_to_data(filterset_class, filter_arg),
            queryset=qs,
            request=info.context,
        )
        if filterset.form.is_valid():
            return filterset.qs
        raise ValidationError(filterset.form.errors.as_json())



"""Module for converting a AdvancedFilterSet class to filter arguments."""

from typing import Any, Dict, List, Optional, Sequence, Type, cast

import graphene
from anytree import Node
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django_filters import Filter
from django_filters.conf import settings as django_settings
from graphene_django.filter.utils import get_model_field
from graphene_django.forms.converter import convert_form_field
from stringcase import pascalcase

from .conf import settings
from .filters import SearchQueryFilter, SearchRankFilter, TrigramFilter
from .filterset import AdvancedFilterSet
from .input_types import (
    SearchQueryFilterInputType,
    SearchRankFilterInputType,
    TrigramFilterInputType,
)


class FilterArgumentsFactory:
    """Factory for creating filter arguments."""

    SPECIAL_FILTER_INPUT_TYPES_FACTORIES = {
        SearchQueryFilter.postfix: lambda: graphene.InputField(
            SearchQueryFilterInputType,
            description='Field for the full text search using '
                        'the `SearchVector` and `SearchQuery` object',
        ),
        SearchRankFilter.postfix: lambda: graphene.InputField(
            SearchRankFilterInputType,
            description='Field for the full text search using the `SearchRank` object',
        ),
        TrigramFilter.postfix: lambda: graphene.InputField(
            TrigramFilterInputType,
            description='Field for the full text search using similarity or distance of trigram',
        ),
    }

    input_object_types: Dict[str, Type[graphene.InputObjectType]] = {}

    def __init__(self, filterset_class: Type[AdvancedFilterSet], input_type_prefix: str) -> None:
        self.filterset_class = filterset_class
        self.input_type_prefix = input_type_prefix
        self.filter_input_type_name = f'{self.input_type_prefix}FilterInputType'

    @property
    def arguments(self) -> Dict[str, graphene.Argument]:
        """Inspect a FilterSet and produce the arguments to pass to a Graphene Field.

        These arguments will be available to filter against in the GraphQL.
        """
        input_object_type = self.input_object_types.get(
            self.filter_input_type_name,
            self.create_filter_input_type(
                self.filterset_to_trees(self.filterset_class),
            ),
        )
        return {
            settings.FILTER_KEY: graphene.Argument(
                input_object_type,
                description='Advanced filter field',
            ),
        }

    def create_filter_input_type(self, roots: List[Node]) -> Type[graphene.InputObjectType]:
        """Create a filter input type from filter set trees."""
        self.input_object_types[self.filter_input_type_name] = cast(
            Type[graphene.InputObjectType],
            type(
                self.filter_input_type_name,
                (graphene.InputObjectType,),
                {
                    **{
                        root.name: self.create_filter_input_subfield(
                            root,
                            self.input_type_prefix,
                            f'`{pascalcase(root.name)}` field',
                        )
                        for root in roots
                    },
                    settings.AND_KEY: graphene.InputField(
                        graphene.List(lambda: self.input_object_types[self.filter_input_type_name]),
                        description='`And` field',
                    ),
                    settings.OR_KEY: graphene.InputField(
                        graphene.List(lambda: self.input_object_types[self.filter_input_type_name]),
                        description='`Or` field',
                    ),
                    settings.NOT_KEY: graphene.InputField(
                        lambda: self.input_object_types[self.filter_input_type_name],
                        description='`Not` field',
                    ),
                },
            ),
        )
        return self.input_object_types[self.filter_input_type_name]

    def create_filter_input_subfield(
        self,
        root: Node,
        prefix: str,
        description: str,
    ) -> graphene.InputField:
        """Create a filter input subfield from a filter set subtree."""
        fields: Dict[str, graphene.InputField] = {}
        if root.name in self.SPECIAL_FILTER_INPUT_TYPES_FACTORIES:
            return self.SPECIAL_FILTER_INPUT_TYPES_FACTORIES[root.name]()
        else:
            for child in root.children:
                if child.height == 0:
                    filter_name = f'{LOOKUP_SEP}'.join(
                        node.name for node in child.path
                        if node.name != django_settings.DEFAULT_LOOKUP_EXPR
                    )
                    fields[child.name] = self.get_field(
                        filter_name,
                        self.filterset_class.base_filters[filter_name],
                    )
                else:
                    fields[child.name] = self.create_filter_input_subfield(
                        child,
                        prefix + pascalcase(root.name),
                        f'`{pascalcase(child.name)}` subfield',
                    )
        return graphene.InputField(
            self.create_input_object_type(
                f'{prefix}{pascalcase(root.name)}FilterInputType',
                fields,
            ),
            description=description,
        )

    @classmethod
    def create_input_object_type(
        cls,
        name: str,
        fields: Dict[str, Any],
    ) -> Type[graphene.InputObjectType]:
        """Create an inheritor for the `InputObjectType` class."""
        if name in cls.input_object_types:
            return cls.input_object_types[name]
        cls.input_object_types[name] = cast(
            Type[graphene.InputObjectType],
            type(
                name,
                (graphene.InputObjectType,),
                fields,
            ),
        )
        return cls.input_object_types[name]

    def get_field(self, name: str, filter_field: Filter) -> graphene.InputField:
        """Return Graphene input field from a filter field.

        It is a partial copy of the `get_filtering_args_from_filterset` function
        from graphene-django.
        https://github.com/graphql-python/graphene-django/blob/caf954861025b9f3d9d3f9c204a7cbbc87352265/graphene_django/filter/utils.py#L11
        """
        model = self.filterset_class._meta.model
        form_field: Optional[models.Field] = None
        filter_type: str = filter_field.lookup_expr
        if name in getattr(self.filterset_class, 'declared_filters'):
            form_field = filter_field.field
            field = convert_form_field(form_field)
        else:
            model_field = get_model_field(model, filter_field.field_name)
            if filter_type != 'isnull' and hasattr(model_field, 'formfield'):
                form_field = model_field.formfield(
                    required=filter_field.extra.get('required', False),
                )
            if not form_field:
                form_field = filter_field.field
            field = convert_form_field(form_field)
        if filter_type in ('in', 'range'):
            field = graphene.List(field.get_type())
        field_type = field.InputField()
        field_type.description = getattr(filter_field, 'label') or \
            f'`{pascalcase(filter_field.lookup_expr)}` lookup'
        return field_type

    @classmethod
    def filterset_to_trees(cls, filterset_class: Type[AdvancedFilterSet]) -> List[Node]:
        """Convert a FilterSet class to trees."""
        trees: List[Node] = []
        for filter_value in filterset_class.base_filters.values():
            values = (*filter_value.field_name.split(LOOKUP_SEP), filter_value.lookup_expr)
            if len(trees) == 0 or not any([cls.try_add_sequence(tree, values) for tree in trees]):
                trees.append(cls.sequence_to_tree(values))
        return trees

    @classmethod
    def try_add_sequence(cls, root: Node, values: Sequence[str]) -> bool:
        """Try to add a sequence to a tree.

        Return a flag indicating whether the mutation was made.
        """
        if root.name == values[0]:
            for child in root.children:
                is_mutated = cls.try_add_sequence(child, values[1:])
                if is_mutated:
                    return True
            root.children = (*root.children, cls.sequence_to_tree(values[1:]))
            return True
        else:
            return False

    @staticmethod
    def sequence_to_tree(values: Sequence[str]) -> Node:
        """Convert a sequence to a tree."""
        node: Optional[Node] = None
        for value in values:
            node = Node(name=value, parent=node)
        return node.root



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphene_django_filter']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3,<4',
 'anytree>=2.8.0,<3.0.0',
 'django-filter>=21.1,<22.0',
 'graphene-django>=2.15.0,<3.0.0',
 'graphene==2.1.9',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'stringcase>=1.2.0,<2.0.0',
 'wrapt>=1.14.0,<2.0.0']

setup_kwargs = {
    'name': 'graphene-django-filter',
    'version': '0.6.4',
    'description': 'Advanced filters for Graphene',
    'long_description': '# Graphene-Django-Filter\n[![CI](https://github.com/devind-team/graphene-django-filter/workflows/CI/badge.svg)](https://github.com/devind-team/graphene-django-filter/actions)\n[![Coverage Status](https://coveralls.io/repos/github/devind-team/graphene-django-filter/badge.svg?branch=main)](https://coveralls.io/github/devind-team/graphene-django-filter?branch=main)\n[![PyPI version](https://badge.fury.io/py/graphene-django-filter.svg)](https://badge.fury.io/py/graphene-django-filter)\n[![License: MIT](https://img.shields.io/badge/License-MIT-success.svg)](https://opensource.org/licenses/MIT)\n\nThis package contains advanced filters for [graphene-django](https://github.com/graphql-python/graphene-django).\nThe standard filtering feature in graphene-django relies on the [django-filter](https://github.com/carltongibson/django-filter)\nlibrary and therefore provides the flat API without the ability to use logical operators such as\n`and`, `or` and `not`. This library makes the API nested and adds logical expressions by extension\nof the `DjangoFilterConnectionField` field and the `FilterSet` class.\nAlso, the library provides some other convenient filtering features.\n\n# Installation\n```shell\n# pip\npip install graphene-django-filter\n# poetry\npoetry add graphene-django-filter\n```\n\n# Requirements\n* Python (3.7, 3.8, 3.9, 3.10)\n* Graphene-Django (2.15)\n\n# Features\n\n## Nested API with the ability to use logical operators\nTo use, simply replace all `DjangoFilterConnectionField` fields with\n`AdvancedDjangoFilterConnectionField` fields in your queries.\nAlso,if you create custom FilterSets, replace the inheritance from the `FilterSet` class\nwith the inheritance from the `AdvancedFilterSet` class.\nFor example, the following task query exposes the old flat API.\n```python\nimport graphene\nfrom django_filters import FilterSet\nfrom graphene_django import DjangoObjectType\nfrom graphene_django.filter import DjangoFilterConnectionField\n\nclass TaskFilter(FilterSet):\n    class Meta:\n        model = Task\n        fields = {\n            \'name\': (\'exact\', \'contains\'),\n            \'user__email\': (\'exact\', \'contains\'),\n            \'user__first_name\': (\'exact\', \'contains\'),\n            \'user__last_name\': (\'exact\', \'contains\'),\n        }\n\nclass UserType(DjangoObjectType):\n    class Meta:\n        model = User\n        interfaces = (graphene.relay.Node,)\n        fields = \'__all__\'\n\nclass TaskType(DjangoObjectType):\n    user = graphene.Field(UserType)\n\n    class Meta:\n        model = Task\n        interfaces = (graphene.relay.Node,)\n        fields = \'__all__\'\n        filterset_class = TaskFilter\n\nclass Query(graphene.ObjectType):\n    tasks = DjangoFilterConnectionField(TaskType)\n```\nThe flat API in which all filters are applied using the `and` operator looks like this.\n```graphql\n{\n  tasks(\n    name_Contains: "important"\n    user_Email_Contains: "john"\n    user_FirstName: "John"\n    user_LastName: "Dou"\n  ){\n    edges {\n      node {\n        id\n        name\n      }\n    }\n  }\n}\n```\nAfter replacing the field class with the `AdvancedDjangoFilterConnectionField`\nand the `FilterSet` class with the `AdvancedFilterSet`\nthe API becomes nested with support for logical expressions.\n```python\nimport graphene\nfrom graphene_django_filter import AdvancedDjangoFilterConnectionField, AdvancedFilterSet\n\nclass TaskFilter(AdvancedFilterSet):\n    class Meta:\n        model = Task\n        fields = {\n            \'name\': (\'exact\', \'contains\'),\n            \'user__email\': (\'exact\', \'contains\'),\n            \'user__first_name\': (\'exact\', \'contains\'),\n            \'user__last_name\': (\'exact\', \'contains\'),\n        }\n\nclass Query(graphene.ObjectType):\n    tasks = AdvancedDjangoFilterConnectionField(TaskType)\n```\nFor example, the following query returns tasks which names contain the word "important"\nor the user\'s email address contains the word "john" and the user\'s last name is "Dou"\nand the first name is not "John".\nNote that the operators are applied to lookups\nsuch as `contains`, `exact`, etc. at the last level of nesting.\n```graphql\n{\n  tasks(\n    filter: {\n      or: [\n        {name: {contains: "important"}}\n        {\n            and: [\n              {user: {email: {contains: "john"}}}\n              {user: {lastName: {exact: "Dou"}}}\n            ]\n        }\n      ]\n      not: {\n        user: {firstName: {exact: "John"}}\n      }\n    }\n  ) {\n    edges {\n      node {\n        id\n        name\n      }\n    }\n  }\n}\n```\nThe same result can be achieved with an alternative query structure\nbecause within the same object the `and` operator is always used.\n```graphql\n{\n  tasks(\n    filter: {\n      or: [\n        {name: {contains: "important"}}\n        {\n          user: {\n            email: {contains: "john"}\n            lastName: {exact: "Dou"}\n          }\n        }\n      ]\n      not: {\n        user: {firstName: {exact: "John"}}\n      }\n    }\n  ){\n    edges {\n      node {\n        id\n        name\n      }\n    }\n  }\n}\n```\nThe filter input type has the following structure.\n```graphql\ninput FilterInputType {\n  and: [FilterInputType]\n  or: [FilterInputType]\n  not: FilterInputType\n  ...FieldLookups\n}\n```\nFor more examples, see [tests](https://github.com/devind-team/graphene-django-filter/blob/06ed0af8def8a4378b4c65a5d137ef17b6176cab/tests/test_queries_execution.py#L23).\n\n## Full text search\nDjango provides the [API](https://docs.djangoproject.com/en/3.2/ref/contrib/postgres/search/)\nfor PostgreSQL full text search. Graphene-Django-Filter inject this API into the GraphQL filter API.\nTo use, add `full_text_search` lookup to fields for which you want to enable full text search.\nFor example, the following type has full text search for\n`first_name` and `last_name` fields.\n```python\nimport graphene\nfrom graphene_django import DjangoObjectType\nfrom graphene_django_filter import AdvancedDjangoFilterConnectionField\n\nclass UserType(DjangoObjectType):\n    class Meta:\n        model = User\n        interfaces = (graphene.relay.Node,)\n        fields = \'__all__\'\n        filter_fields = {\n            \'email\': (\'exact\', \'startswith\', \'contains\'),\n            \'first_name\': (\'exact\', \'contains\', \'full_text_search\'),\n            \'last_name\': (\'exact\', \'contains\', \'full_text_search\'),\n        }\n\nclass Query(graphene.ObjectType):\n    users = AdvancedDjangoFilterConnectionField(UserType)\n```\nSince this feature belongs to the AdvancedFilterSet,\nit can be used in a custom FilterSet.\nThe following example will work exactly like the previous one.\n```python\nimport graphene\nfrom graphene_django import DjangoObjectType\nfrom graphene_django_filter import AdvancedDjangoFilterConnectionField, AdvancedFilterSet\n\nclass UserFilter(AdvancedFilterSet):\n    class Meta:\n        model = User\n        fields = {\n            \'email\': (\'exact\', \'startswith\', \'contains\'),\n            \'first_name\': (\'exact\', \'contains\', \'full_text_search\'),\n            \'last_name\': (\'exact\', \'contains\', \'full_text_search\'),\n        }\n\nclass UserType(DjangoObjectType):\n    class Meta:\n        model = User\n        interfaces = (graphene.relay.Node,)\n        fields = \'__all__\'\n        filterset_class = UserFilter\n\nclass Query(graphene.ObjectType):\n    users = AdvancedDjangoFilterConnectionField(UserType)\n```\nFull text search API includes SearchQuery, SearchRank, and Trigram filters.\nSearchQuery and SearchRank filters are at the top level.\nIf some field has been enabled for full text search then it can be included in the field array.\nThe following queries show an example of using the SearchQuery and SearchRank filters.\n```graphql\n{\n  users(\n    filter: {\n      searchQuery: {\n        vector: {\n          fields: ["first_name"]\n        }\n        query: {\n          or: [\n            {value: "Bob"}\n            {value: "Alice"}\n          ]\n        }\n      }\n    }\n  ){\n    edges {\n      node {\n        id\n        firstName\n        lastName  \n      }\n    }\n  }\n}\n```\n```graphql\n{\n  users(\n    filter: {\n      searchRank: {\n        vector: {fields: ["first_name", "last_name"]}\n        query: {value: "John Dou"}\n        lookups: {gte: 0.5}\n      }\n    }\n  ){\n    edges {\n      node {\n        id\n        firstName\n        lastName  \n      }\n    }\n  }\n}\n```\nTrigram filter belongs to the corresponding field.\nThe following query shows an example of using the Trigram filter.\n```graphql\n{\n  users(\n    filter: {\n      firstName: {\n        trigram: {\n          value: "john"\n          lookups: {gte: 0.85}\n        }\n      }\n    }\n  ){\n    edges {\n      node {\n        id\n        firstName\n        lastName  \n      }\n    }\n  }\n}\n```\nInput types have the following structure.\n```graphql\ninput SearchConfigInputType {\n  value: String!\n  isField: Boolean\n}\nenum SearchVectorWeight {\n  A\n  B\n  C\n  D\n}\ninput SearchVectorInputType {\n  fields: [String!]!\n  config: SearchConfigInputType\n  weight: SearchVectorWeight\n}\nenum SearchQueryType {\n  PLAIN\n  PHRASE\n  RAW\n  WEBSEARCH\n}\ninput SearchQueryInputType {\n  value: String\n  config: SearchConfigInputType\n  and: [SearchQueryInputType]\n  or: [SearchQueryInputType]\n  not: SearchQueryInputType\n}\ninput SearchQueryFilterInputType {\n  vector: SearchVectorInputType!\n  query: SearchQueryInputType!\n}\ninput FloatLookupsInputType {\n  exact: Float\n  gt: Float\n  gte: Float\n  lt: Float\n  lte: Float\n}\ninput SearchRankWeightsInputType {\n  D: Float\n  C: Float\n  B: Float\n  A: Float\n}\ninput SearchRankFilterInputType {\n  vector: SearchVectorInputType!\n  query: SearchQueryInputType!\n  lookups: FloatLookupsInputType!\n  weights: SearchRankWeightsInputType\n  coverDensity: Boolean\n  normalization: Int\n}\nenum TrigramSearchKind {\n  SIMILARITY\n  DISTANCE\n}\ninput TrigramFilterInputType {\n  kind: TrigramSearchKind\n  lookups: FloatLookupsInputType!\n  value: String!\n}\n```\nFor more examples, see [tests](https://github.com/devind-team/graphene-django-filter/blob/06ed0af8def8a4378b4c65a5d137ef17b6176cab/tests/test_queries_execution.py#L134).\n\n## Settings\nThe library can be customised using settings.\nTo add settings, create a dictionary\nwith name `GRAPHENE_DJANGO_FILTER` in the project’s `settings.py`.\nThe default settings are as follows.\n```python\nGRAPHENE_DJANGO_FILTER = {\n    \'FILTER_KEY\': \'filter\',\n    \'AND_KEY\': \'and\',\n    \'OR_KEY\': \'or\',\n    \'NOT_KEY\': \'not\',\n}\n```\nTo read the settings, import them from the `conf` module.\n```python\nfrom graphene_django_filter.conf import settings\n\nprint(settings.FILTER_KEY)\n```\nThe `settings` object also includes fixed settings, which depend on the user\'s environment.\n`IS_POSTGRESQL` determinate that current database is PostgreSQL\nand `HAS_TRIGRAM_EXTENSION` that `pg_trgm` extension is installed.\n',
    'author': 'devind-team',
    'author_email': 'team@devind.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/devind-team/graphene-django-filter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""InputObjectType classes for special lookups."""

from typing import Type, cast

import graphene

from .conf import settings


class SearchConfigInputType(graphene.InputObjectType):
    """Input type for the `SearchVector` or `SearchQuery` object config."""

    value = graphene.String(
        required=True,
        description='`SearchVector` or `SearchQuery` object config value',
    )
    is_field = graphene.Boolean(
        default_value=False,
        description='Whether to wrap the value with the F object',
    )


class SearchVectorWeight(graphene.Enum):
    """Weight of the `SearchVector` object."""

    A = 'A'
    B = 'B'
    C = 'C'
    D = 'D'


class SearchVectorInputType(graphene.InputObjectType):
    """Input type for creating the `SearchVector` object."""

    fields = graphene.InputField(
        graphene.List(graphene.NonNull(graphene.String)),
        required=True,
        description='Field names of vector',
    )
    config = graphene.InputField(SearchConfigInputType, description='Vector config'),
    weight = graphene.InputField(SearchVectorWeight, description='Vector weight')


class SearchQueryType(graphene.Enum):
    """Search type of the `SearchQuery` object."""

    PLAIN = 'plain'
    PHRASE = 'phrase'
    RAW = 'raw'
    WEBSEARCH = 'websearch'


def create_search_query_input_type() -> Type[graphene.InputObjectType]:
    """Return input type for creating the `SearchQuery` object."""
    search_query_input_type = cast(
        Type[graphene.InputObjectType],
        type(
            'SearchQueryInputType',
            (graphene.InputObjectType,),
            {
                '__doc__': 'Input type for creating the `SearchQuery` object.',
                'value': graphene.String(description='Query value'),
                'config': graphene.InputField(SearchConfigInputType, description='Query config'),
                settings.AND_KEY: graphene.InputField(
                    graphene.List(graphene.NonNull(lambda: search_query_input_type)),
                    description='`And` field',
                ),
                settings.OR_KEY: graphene.InputField(
                    graphene.List(graphene.NonNull(lambda: search_query_input_type)),
                    description='`Or` field',
                ),
                settings.NOT_KEY: graphene.InputField(
                    graphene.List(graphene.NonNull(lambda: search_query_input_type)),
                    description='`Not` field',
                ),
            },
        ),
    )
    return search_query_input_type


SearchQueryInputType = create_search_query_input_type()


class SearchQueryFilterInputType(graphene.InputObjectType):
    """Input type for the full text search using the `SearchVector` and `SearchQuery` object."""

    vector = graphene.InputField(SearchVectorInputType, required=True, description='Search vector')
    query = graphene.InputField(SearchQueryInputType, required=True, description='Search query')


class FloatLookupsInputType(graphene.InputObjectType):
    """Input type for float lookups."""

    exact = graphene.Float(description='Is exact')
    gt = graphene.Float(description='Is greater than')
    gte = graphene.Float(description='Is greater than or equal to')
    lt = graphene.Float(description='Is less than')
    lte = graphene.Float(description='Is less than or equal to')


class SearchRankWeightsInputType(graphene.InputObjectType):
    """`SearchRank` object weights.

    Default values are set according to the documentation.
    https://docs.djangoproject.com/en/3.2/ref/contrib/postgres/search/#weighting-queries
    """

    D = graphene.Float(default_value=0.1, description='D letter')
    C = graphene.Float(default_value=0.2, description='C letter')
    B = graphene.Float(default_value=0.4, description='B letter')
    A = graphene.Float(default_value=1.0, description='A letter')


class SearchRankFilterInputType(graphene.InputObjectType):
    """Input type for the full text search using the `SearchRank` object."""

    vector = graphene.InputField(SearchVectorInputType, required=True, description='Search vector')
    query = graphene.InputField(SearchQueryInputType, required=True, description='Search query')
    lookups = graphene.InputField(
        FloatLookupsInputType,
        required=True,
        description='Available lookups',
    )
    weights = graphene.InputField(SearchRankWeightsInputType, description='Search rank weights')
    cover_density = graphene.Boolean(
        default_value=False,
        description='Whether to include coverage density ranking',
    )
    normalization = graphene.Int(description='Search rank normalization')


class TrigramSearchKind(graphene.Enum):
    """Type of the search using trigrams."""

    SIMILARITY = 'similarity'
    DISTANCE = 'distance'


class TrigramFilterInputType(graphene.InputObjectType):
    """Input type for the full text search using similarity or distance of trigram."""

    kind = graphene.InputField(
        TrigramSearchKind,
        default_value=TrigramSearchKind.SIMILARITY,
        description='Type of the search using trigrams',
    )
    lookups = graphene.InputField(
        FloatLookupsInputType,
        required=True,
        description='Available lookups',
    )
    value = graphene.String(required=True, description='Search value')
